package beagle.bgtheory.LIA.cooper

import beagle._
import fol._
import util._
import datastructures.Clause
import bgtheory._
import LIA._

object SolverClauses extends LIASolver {
  val name = "cooper-clauses"
  def hasSolverLiterals(cl: Clause) = true
  def toSolverLiterals(cl: Clause) = cl
  def check(cls: Iterable[Clause]) =
    try {
      if (cls forall  { _.isGround }) 
	toSolverResult(isConsistent(cls))
      else 
	toSolverResult(LIA.cooper.QE((cls map { _.toFormula }).toAnd, false, cls flatMap { _.symConsts }) != FalseAtom)
    } catch {
      // todo: is this fallback really a good idea? Why are we doing this?
      // JOSH: this is in case the coefficients of some variable are too large to
      // form the LCM. Since BNB doesn't need this it can usually work when
      // cooper doesn't.
      // Ideally we would just be able to cleverly cast to a Long or BigInt to avoid
      // this issue while still using cooper.
      case IntOverflowException => throw InternalError("Integer overflow in Cooper")
    }

  val needsUnitClauses = false
}

object SolverUnits extends LIASolver {
  //TODO- can we maintain a state here?
  //e.g. lastcls = C
  //next could be C & L
  val name = "cooper-units"
  def hasSolverLiterals(cl: Clause) = true
  def toSolverLiterals(cl: Clause) = cl
  // def isConsistent(f: Formula) = cooper.isConsistent(f)
  def check(cls: Iterable[Clause]) = 
      toSolverResult(isConsistent(cls))
    // try {
    //   toSolverResult(cooper.isConsistent(cls))
    // } catch {
    //   case e: InternalError => BNBSolverUnits.check(cls)
    // }

  val needsUnitClauses = true
}

