package beagle.cnfconversion

import beagle._
import fol._
import Signature._
import util._
import rules._

/*
 * Naive Skolemization
 */

object skolemization {

  def skolemize(g: Formula): Formula = {
    // Assume all quantifiers to be eliminated have been pulled out 
    // If full is false then do not skolemize background formulas. Rationale is that QE takes care of quantifiers for BG formulas 
    // println("skolemize on " + g)
    // body of skolemize
    g match {
      // Simple cases to collapse quantifiers
      case Forall(xs, Forall(ys, f)) ⇒ skolemize(Forall(xs ::: ys, f))
      case Exists(xs, Exists(ys, f)) ⇒ skolemize(Exists(xs ::: ys, f))

      case Forall(xs, Exists(ys, f)) ⇒ {
        var sigma = Subst.empty // The skolemization substitution, for all vars in ys
        ys foreach { sigma += mkSkoSubst(xs, _) }
        skolemize(Forall(xs, sigma(f)))
      }
      // Case of one single row of exists-quantifier, without forall in front.
      case Exists(ys, f) ⇒ {
        var sigma = Subst.empty // The skolemization substitution, for all vars in ys
        ys foreach { sigma += mkSkoSubst(List.empty, _) }
        skolemize(sigma(f))
      }
      // skolemize(Forall(List.empty, Exists(ys, f)))
      case _ ⇒ g // only universal quantifiers, if any 
    }
  }

  // Make a Skolemizing substitution for the variable x in context of leading universally quantified variables xs 
  // As a side effect add the skolem function to Sigma
  def mkSkoSubst(xs: List[Var], y: Var) = {
    val skoFun = "#skF_" + skfCtr.next() // #skf
    val skoTermArity = Arity((xs.sortsOf), y.sort)
    // Figure out if we can make a background constants
    val skoTermKind =
      if ((y.sort.kind == BG) && xs.isEmpty && flags.paramsOpSet.value == "BG")
        BG
      else
        FG
    //val skoOp = Operator(skoTermKind, skoFun, skoTermArity)
    val skoOp = new SkolemOp(skoTermKind, skoFun, skoTermArity)
    // todo: not added - is this the Right Thing?
    Sigma += skoOp
    (y -> NonDomElemFunTerm(skoOp, xs)) // Technically, this is not a Subst but never mind
  }

  val skfCtr = new Counter

}
